#include "Camera.hpp"

Camera::Camera(Void) : _camera(null), _bitmap(null), _previewable(false), _recordable(false), _spp(null)
{
    _mediaData.clsData = MMD_FILE_NAME;
    _mediaData.pData = "camera.jpg";
    _mediaData.dwSize = 0;
    _mimeType = "image/jpeg";
    return;
}

Camera::~Camera(Void)
{
    // Destructor
    Terminate();
    return;
}

//
// rect         : Preview display area
// spp          : Pointer to the external callback function
// reference    : Pointer to data passed to external callback function
//
SFCError Camera::Initialize(SFXRectangleConstRef rect, NotifySPP spp, VoidPtr reference)
{

    SFCError error(SFERR_NO_ERROR);
    // Make bitmap for photo image
    _bitmap = SFXGraphics::CreateBitmap(rect.GetSize());
    if (_bitmap != null) {
        // Preparations for camera processing
        error = Resume();
        if (error == SFERR_NO_ERROR) {
            _rect = rect;
            _spp = spp;
            _reference = reference;
        }
    } else {
        // Bitmap cannot be made because of insufficient memory
        error = SFERR_NO_MEMORY;
    }
    if (error != SFERR_NO_ERROR) {
        Terminate();
    }
    return error;
}

//
// Start preview.
//
SFCError Camera::Preview(Void)
{
    SFCError error(SFERR_NO_ERROR);
    if (_camera != null) {
        _status = STATUS_PREVIEW;
        _previewable = false;
        _recordable = false;
        // Boot up the camera for previewing
        error = _camera->Preview();
    } else {
        error = SFERR_FAILED;
    }
    return error;
}

//
// Take a picture in the snapshot mode.
//
SFCError Camera::Record(Void)
{
    SFCError error(SFERR_NO_ERROR);

    if (_camera != null) {
        _previewable = false;
        _recordable = false;

        _status = STATUS_RECORD;
        // Ending the preview
        // An event handler is called when the preview actually ends
        // and takes a picture in snapshot mode
        error = _camera->Stop();
    }
    else {
        error = SFERR_FAILED;
    }
    return error;
}

//
// Camera termination.
//
Void Camera::Suspend(Void)
{
    if (_camera != null) {
        _previewable = false;
        _recordable = false;
        _camera->Stop();
        _camera.Release();
    }
    return;
}

//
// Basic camera initialization.
//
SFCError Camera::Resume(Void)
{
    SFCError error(SFERR_NO_ERROR);
    // Get a camera instance
    if ((_camera = SFBCamera::NewInstance()) != null) {
        _previewable = true;
        _recordable = false;
        // Register internal callback
        error = _camera->RegisterNotify(OnCameraSHP, this);
        if (error == SFERR_NO_ERROR) {
            AEESize size;
            size.cx = PICTURE_WIDTH;
            size.cy = PICTURE_HEIGHT;
            error = _camera->SetSize(&size);
            if (error == SFERR_NO_ERROR) {
                // Set the size of photo image
                error = _camera->SetDisplaySize(&size);
            }
        }
    } else {
        // Camera instance cannot be gotten because of insufficient memory
        error = SFERR_NO_MEMORY;
    }
    if (error != SFERR_NO_ERROR) {
        Suspend();
    }
    return error;
}

//
// Once the preview is finished, and the source has been used, release it.
//
Void Camera::Terminate(Void)
{
    Suspend();
    if (_bitmap != null) {
        // The bitmap is released
        _bitmap.Release();
    }
    return;
}

//
// Processing after taking a snapshot picture.
//
Void Camera::AfterSnapshot(Void)
{
    SFBBitmapSmp bitmap;
    // The image of the camera as a bitmap
    if (_camera->GetFrame(&bitmap) == SFERR_NO_ERROR) {
        // Display the size of camera image by designating size
        _bitmap->BltIn(SFXRectangle(0, 0, _rect.GetSize()), bitmap, _rect.GetOrigin());
    }
    _previewable = true;
    _spp(_status, SFERR_NO_ERROR, _reference);
}

//
// Internal Callback Handler (dummy).
//
Void Camera::OnCameraSHP(VoidPtr reference, AEECameraNotify* notify)
{
    // Call the body of Internal Callback Handler
    static_cast<CameraPtr>(reference)->OnCamera(notify); 
    return;
}

//
// Internal Callback Handler (body).
//
Void Camera::OnCamera(AEECameraNotify *notify)
{
    switch (notify->nStatus) {
        case CAM_STATUS_START:
            // The boot up of the camera was completed
            if (_status == STATUS_PREVIEW) {
                _recordable = true;
                // Call external callback function
                _spp(_status, SFERR_NO_ERROR, _reference);
            }
            break;
        case CAM_STATUS_FRAME:
            // A new frame was obtained
            if (_status == STATUS_PREVIEW) {
                SFBBitmapSmp bitmap;
                // The photo image of the camera is obtained as a bitmap
                if (_camera->GetFrame(&bitmap) == SFERR_NO_ERROR) {
                    // Clip displayed area from photo image
                    _bitmap->BltIn(SFXRectangle(0, 0, _rect.GetSize()), bitmap, _rect.GetOrigin());
                }
                // Call external callback function
                _spp(STATUS_FRAME, SFERR_NO_ERROR, _reference);

                SFXGraphicsPtr pg = SFXGraphics::GetInstance();
                // The obtained image is displayed on the screen
                pg->BitBlt(_rect, _bitmap, SFXGrid(0, 0));
                // Update screen
                pg->Update();
            }
            break;
        case CAM_STATUS_DONE:
            // Processing was completed
            if (_status == STATUS_RECORD) {
                _status = STATUS_ENCODE;
                // Set delay encode mode
                _camera->DeferEncode(true);
                _camera->SetMediaData(&_mediaData, _mimeType);
                // Take a picture in the snapshot mode
                _camera->RecordSnapshot();
            } else if (_status == STATUS_ENCODE) {
                AfterSnapshot();
            }
            break;
        case CAM_STATUS_FAIL:
        case CAM_STATUS_ABORT:
            //Call external callback function
            _spp(_status, EFAILED, _reference);
            break;
    }
    return;
}

//
// Get current brightness.
//
SInt16 Camera::GetBrightness(Void)
{
    SInt16 result(0);
    if (_camera != null) {
        SInt32 value;
        AEEParmInfo info;
        // Get brightness
        if (_camera->GetParm(CAM_PARM_BRIGHTNESS, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {
            result = (SInt16)value;
        }
    }
    return result;
}

//
// Get range of brightness.
//
Bool Camera::GetBrightnessRange(SInt16Ptr minimum, SInt16Ptr maximum)
{
    Bool result(false);
    *minimum = *maximum = 0;
    if (_camera != null) {
        Bool support;
        if (_camera->IsBrightness(&support) == SFERR_NO_ERROR) {
            if (support) {
                SInt32 value;
                AEEParmInfo info;
                // Get range
                if (_camera->GetParm(CAM_PARM_BRIGHTNESS, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {
                    *minimum = (SInt16)info.nMin;
                    *maximum = (SInt16)info.nMax;
                    result = true;
                }
            }
        }
    }
    return result;
}

//
// Set brightness.
//
Void Camera::SetBrightness(SInt16 value)
{
    if (_camera != null) {
        if (_recordable) {
            // Set brightness
            _camera->SetBrightness(value);
        }
    }
}

//
// Get current zoom value.
//
SInt16 Camera::GetZoom(Void)
{
    SInt16 result(0);
    if (_camera != null) {
        SInt32 value;
        AEEParmInfo info;
         // Get zoom value
        if (_camera->GetParm(CAM_PARM_ZOOM, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {
            result = (SInt16)value;
        }
    }
    return result;
}

//
// Get range of zoom value.
//
Bool Camera::GetZoomRange(SInt16Ptr minimum, SInt16Ptr maximum)
{
    Bool result(false);
    *minimum = *maximum = 0;
    if (_camera != null) {
        Bool support;
        if (_camera->IsZoom(&support) == SFERR_NO_ERROR) {
            if (support) {
                SInt32 value;
                AEEParmInfo info;
                // Get range
                if (_camera->GetParm(CAM_PARM_ZOOM, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {
                    *minimum = (SInt16)info.nMin;
                    *maximum = (SInt16)info.nMax;
                    result = true;
                }
            }
        }
    }
    return result;
}

//
// Set zoom value.
//
Void Camera::SetZoom(SInt16 value)
{
    if (_camera != null) {
        if (_recordable) {
            // Setting zoom value
            _camera->SetZoom(value);
        }
    }
}
